//! Errors for arg parsing
use super::*;
use flags::*;
use std::{
    error,
    fmt,
};

/// Error without context
#[derive(Debug)]
pub struct Error
{
    pub(super) flag: Kind<'static>,
    pub(super) arg: (usize, String),
    pub(super) desc: Option<String>,
    pub(super) from: eyre::Report,
}

/// An argument parsing error with context
#[repr(transparent)]
#[derive(Debug)]
pub struct ContextualError(Error);

#[derive(Debug, Clone, Copy)]
pub struct ErrorContext<'a>
{
    /// The error-throwing flag
    pub flag: &'a Kind<'static>,
    /// The argument string
    pub arg: &'a String,
    /// The index of the argument string
    pub arg_idx: &'a usize,
    /// Message from the failing flag callback, if any
    pub desc: Option<&'a String>,
    /// The error report
    pub from: &'a eyre::Report,
}

impl ContextualError
{
    /// The inner error
    #[inline] pub fn inner(&self) -> &Error
    {
	&self.0
    }
    /// Strips the context and returns the inner error
    #[inline] fn into_inner(self) -> Error
    {
	self.0
    }
    /// The context of this error
    pub fn context(&self) -> ErrorContext<'_>
    {
	ErrorContext {
	    flag: &self.0.flag,
	    arg: &self.0.arg.1,
	    arg_idx: &self.0.arg.0,
	    desc: self.0.desc.as_ref(),
	    from: &self.0.from
	}
    }

    /// Which argument was the error thrown for
    pub fn what(&self) -> (usize, &str)
    {
	(self.0.arg.0, &self.0.arg.1[..])
    }

    /// Where did the parsing fail?
    pub fn which(&self) -> Kind<'_>
    {
	self.0.flag.as_ref()
    }

    /// The message reported by the failing callback, if any
    pub fn message(&self) -> Option<&str>
    {
	self.0.desc.as_ref().map(|x| &x[..])
    }
}

impl error::Error for Error
{
    fn source(&self) -> Option<&(dyn error::Error + 'static)> {
	self.from.source()
    }
}

impl fmt::Display for Error
{
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	writeln!(f, "Failed to parse args")
    }
}

impl fmt::Display for ContextualError
{
    #[inline] fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result
    {
	self.0.fmt(f)
    }
}


impl Error
{
    /// Consume into an `eyre::Report`.
    pub(super) fn into_report(self) -> eyre::Report
    {
	let Error {flag, desc, from ,arg: (idx, arg)} = self;
	let  err = Err::<!, _>(from)
	//.wrap_err(eyre!("Failed to parse args"))
	    .with_section(move || flag.to_string().header("While attempting to parse for"))
	    .with_note(move || idx.to_string().header("Argument index was"))
	    .with_note(move || format!("{:?}", arg).header("Argument was"));
	if let Some(desc) = desc {
	    err.with_suggestion(move || desc)
	} else {
	    err
	}.unwrap_err()
    }

    /// Add context to this error
    pub fn with_context(self) -> ContextualError
    {
	ContextualError(Self {
	    flag: self.flag.clone(),
	    arg: self.arg.clone(),
	    desc: self.desc.clone(),
	    from: self.into_report(),
	})
    }
}

impl From<Error> for ContextualError
{
    #[inline] fn from(from: Error) -> Self
    {
	from.with_context()
    }
}

impl From<ContextualError> for eyre::Report
{
    #[inline] fn from(from: ContextualError) -> Self
    {
	from.0.from
    }
}

